from __future__ import unicode_literals
import os
from mediachain.datastore import get_db
from mediachain.datastore.utils import ref_base58, bytes_for_object, \
    object_for_bytes
from mediachain.datastore.data_objects import MultihashReference
from mediachain.utils.file import mkdir_p, system_cache_dir


class InMemoryCache(object):
    def __init__(self):
        self.cache = dict()

    def get(self, key):
        return self.cache.get(key, None)

    def put(self, key, value):
        self.cache[key] = value


def open_db(path):
    try:
        import rocksdb
        opts = rocksdb.Options(create_if_missing=True)
        return rocksdb.DB(path, opts)
    except ImportError:
        print('using in-memory blockchain cache. please install '
              'rocksdb to enable persistent cache')
        return InMemoryCache()


__SHARED_BLOCK_CACHE = None


def get_block_cache():
    global __SHARED_BLOCK_CACHE
    if __SHARED_BLOCK_CACHE is None:
        __SHARED_BLOCK_CACHE = BlockCache()
    return __SHARED_BLOCK_CACHE


class BlockCache(object):
    """
    A read-through cache for mediachain journal blocks.
    When a block is read, it gets written to a local RocksDB cache.
    """
    def __init__(self, datastore=None, cache_path=None):
        if datastore is None:
            datastore = get_db()
        if cache_path is None:
            cache_dir = system_cache_dir('io.mediachain')
            mkdir_p(cache_dir)
            cache_path = os.path.join(cache_dir, 'block-cache.db')

        self.datastore = datastore
        self.cache_path = cache_path
        self.cache_db = open_db(cache_path)

    def put(self, block, ref=None):
        block_bytes = bytes_for_object(block)
        if ref is None:
            ref = MultihashReference.for_content(block_bytes)
        ref = ref_base58(ref)
        self.cache_db.put(bytes(ref), bytes(block_bytes))

    def get(self, block_ref):
        block_ref = bytes(ref_base58(block_ref))

        cached = self.cache_db.get(block_ref)
        if cached is not None:
            return object_for_bytes(cached)

        block = self.datastore.get(block_ref)
        self.put(block, block_ref)
        return object_for_bytes(block)



import base58
import json
from pygments import highlight
from pygments.lexers.data import JsonLexer
from pygments.formatters.terminal import TerminalFormatter

def dump(obj):
    j = json.dumps(stringify_refs(obj), indent=2)
    print highlight(j, JsonLexer(), TerminalFormatter())


def stringify_refs(obj):
    res = {}
    for k, v in obj.iteritems():
        if isinstance(v, dict):
            v = stringify_refs(v)
        if k == u'@link':
            v = base58.b58encode(v)
        res[k] = v
    return res




from mediachain.datastore.ipfs import get_ipfs_datastore
import sys
import os
import shutil
from os.path import expanduser, dirname, join

class ChDir(object):
    """
    Step into a directory temporarily
    """
    def __init__(self, path):
        self.old_dir = os.getcwd()
        self.new_dir = path

    def __enter__(self):
        os.chdir(self.new_dir)

    def __exit__(self, *args):
        os.chdir(self.old_dir)

def get_translator(translator_id):
	try:
		name, version = translator_id.split('@')
	except ValueError:
		raise LookupError(
			"Bad translator id `{}`, must be `name@multihash` format".format(translator_id)
		)

	ipfs = get_ipfs_datastore() # FIXME: memoize this

	base_path = join(expanduser('~'), '.mediachain')
	path = join(base_path, '_mediachain', 'translation')
	if not os.path.exists(path):
	    os.makedirs(path)
	    # print join(path, '__init__.py')
	    open(join(path, '__init__.py'), 'a').close()
	    # print join(dirname(path), '__init__.py')
	    open(join(dirname(path), '__init__.py'), 'a').close()

	if not os.path.exists(path):
	    os.makedirs(path)

	with ChDir(path):
		shutil.rmtree(name, ignore_errors=True)
		translator = ipfs.client.get(version) # FIXME: timeout, error handling
		os.rename(version, name) # ipfsApi doesn't support -o

	sys.path.insert(1, base_path)

	full_name = '_mediachain.translation.' + name + '.translator'
	translator_module = __import__(full_name, globals(), locals(), [name])
	translator = getattr(translator_module, name.capitalize())

	return translator



import functools
from grpc.beta import implementations
from grpc.beta.interfaces import StatusCode
from grpc.framework.interfaces.face.face import AbortionError
from mediachain.proto import Datastore_pb2  # pylint: disable=no-name-in-module
from mediachain.datastore.data_objects import MultihashReference
from mediachain.datastore.utils import rpc_ref, multihash_ref, \
    bytes_for_object, object_for_bytes
from mediachain.rpc.utils import with_retry

TIMEOUT_SECS = 120

__RPC_STORE_CONFIG = None
__RPC_STORE_INSTANCE = None


def set_rpc_datastore_config(cfg):
    global __RPC_STORE_CONFIG
    __RPC_STORE_CONFIG = cfg


def get_rpc_datastore_config():
    global __RPC_STORE_CONFIG
    return __RPC_STORE_CONFIG


def get_db():
    global __RPC_STORE_INSTANCE
    if __RPC_STORE_INSTANCE is not None:
        return __RPC_STORE_INSTANCE

    cfg = get_rpc_datastore_config()
    try:
        host = cfg['host']
        port = int(cfg['port'])
    except (TypeError, LookupError, ValueError):
        raise RuntimeError('No RPC datastore configuration has been set. ' +
                           'please call set_rpc_datastore_config before ' +
                           'calling get_db')

    __RPC_STORE_INSTANCE = RpcDatastore(host, port)
    return __RPC_STORE_INSTANCE


def close_db():
    """
    Cleanup the global RpcDatastore instance.
    Must be called before program exit, otherwise the connection will never
    be closed, and grpc will hang indefinitely.
    """
    global __RPC_STORE_INSTANCE
    __RPC_STORE_INSTANCE = None


class RpcDatastore(object):
    def __init__(self, host, port):
        channel = implementations.insecure_channel(host, port)
        self.rpc = Datastore_pb2.beta_create_DatastoreService_stub(channel)
        self.put_cache = set()

    def put(self, data_object, timeout=TIMEOUT_SECS):
        put_with_retry = functools.partial(with_retry, self.rpc.put)
        byte_string = bytes_for_object(data_object)
        content_hash = MultihashReference.for_content(byte_string)
        if content_hash.multihash in self.put_cache:
            return content_hash

        req = Datastore_pb2.DataObject(data=byte_string)
        ref = multihash_ref(put_with_retry(req, timeout))
        self.put_cache.add(ref.multihash)
        return ref

    def get(self, ref, timeout=TIMEOUT_SECS):
        ref = rpc_ref(ref)
        get_with_retry = functools.partial(with_retry, self.rpc.get)
        try:
            obj_bytes = get_with_retry(ref, timeout).data
        except AbortionError as e:
            if e.code == StatusCode.NOT_FOUND:
                return None
            raise
        return object_for_bytes(obj_bytes)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: Types.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='Types.proto',
  package='protocol',
  syntax='proto3',
  serialized_pb=_b('\n\x0bTypes.proto\x12\x08protocol\"\'\n\x12MultihashReference\x12\x11\n\treference\x18\x01 \x01(\t\"\x0f\n\rNullReference\"t\n\x0e\x43hainReference\x12-\n\x05\x63hain\x18\x01 \x01(\x0b\x32\x1c.protocol.MultihashReferenceH\x00\x12&\n\x03nil\x18\x02 \x01(\x0b\x32\x17.protocol.NullReferenceH\x00\x42\x0b\n\treferenceB\x1e\n\x1cio.mediachain.protocol.typesb\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_MULTIHASHREFERENCE = _descriptor.Descriptor(
  name='MultihashReference',
  full_name='protocol.MultihashReference',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='reference', full_name='protocol.MultihashReference.reference', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25,
  serialized_end=64,
)


_NULLREFERENCE = _descriptor.Descriptor(
  name='NullReference',
  full_name='protocol.NullReference',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=66,
  serialized_end=81,
)


_CHAINREFERENCE = _descriptor.Descriptor(
  name='ChainReference',
  full_name='protocol.ChainReference',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='chain', full_name='protocol.ChainReference.chain', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='nil', full_name='protocol.ChainReference.nil', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='reference', full_name='protocol.ChainReference.reference',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=83,
  serialized_end=199,
)

_CHAINREFERENCE.fields_by_name['chain'].message_type = _MULTIHASHREFERENCE
_CHAINREFERENCE.fields_by_name['nil'].message_type = _NULLREFERENCE
_CHAINREFERENCE.oneofs_by_name['reference'].fields.append(
  _CHAINREFERENCE.fields_by_name['chain'])
_CHAINREFERENCE.fields_by_name['chain'].containing_oneof = _CHAINREFERENCE.oneofs_by_name['reference']
_CHAINREFERENCE.oneofs_by_name['reference'].fields.append(
  _CHAINREFERENCE.fields_by_name['nil'])
_CHAINREFERENCE.fields_by_name['nil'].containing_oneof = _CHAINREFERENCE.oneofs_by_name['reference']
DESCRIPTOR.message_types_by_name['MultihashReference'] = _MULTIHASHREFERENCE
DESCRIPTOR.message_types_by_name['NullReference'] = _NULLREFERENCE
DESCRIPTOR.message_types_by_name['ChainReference'] = _CHAINREFERENCE

MultihashReference = _reflection.GeneratedProtocolMessageType('MultihashReference', (_message.Message,), dict(
  DESCRIPTOR = _MULTIHASHREFERENCE,
  __module__ = 'Types_pb2'
  # @@protoc_insertion_point(class_scope:protocol.MultihashReference)
  ))
_sym_db.RegisterMessage(MultihashReference)

NullReference = _reflection.GeneratedProtocolMessageType('NullReference', (_message.Message,), dict(
  DESCRIPTOR = _NULLREFERENCE,
  __module__ = 'Types_pb2'
  # @@protoc_insertion_point(class_scope:protocol.NullReference)
  ))
_sym_db.RegisterMessage(NullReference)

ChainReference = _reflection.GeneratedProtocolMessageType('ChainReference', (_message.Message,), dict(
  DESCRIPTOR = _CHAINREFERENCE,
  __module__ = 'Types_pb2'
  # @@protoc_insertion_point(class_scope:protocol.ChainReference)
  ))
_sym_db.RegisterMessage(ChainReference)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n\034io.mediachain.protocol.types'))
import grpc
from grpc.beta import implementations as beta_implementations
from grpc.beta import interfaces as beta_interfaces
from grpc.framework.common import cardinality
from grpc.framework.interfaces.face import utilities as face_utilities
# @@protoc_insertion_point(module_scope)



import os
import platform
import tempfile
import errno


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def system_cache_dir(subdir=None):
    system = platform.system()
    if system == 'Linux':
        cache_dir = os.path.join(os.getenv('HOME'), '.cache')
    elif system == 'Darwin':
        cache_dir = os.path.join(os.getenv('HOME'), 'Library', 'Caches')
    elif system == 'Windows':  # it could happen :)
        cache_dir = os.getenv('LOCALAPPDATA')
    else:
        cache_dir = tempfile.gettempdir()
    if subdir is None:
        return cache_dir
    return os.path.join(cache_dir, subdir)



import time
import random
from grpc.beta.interfaces import StatusCode
from grpc.framework.interfaces.face.face import AbortionError

BACKOFF_COEFF = 50.0
MAX_ATTEMPTS = 10
RETRYABLE_ERRORS = [
    StatusCode.UNAVAILABLE,
    StatusCode.UNKNOWN
]


def with_retry(func, *args, **kwargs):
    try:
        operation = str(func._method)
    except AttributeError:
        try:
            operation = func.__name__
        except AttributeError:
            operation = 'UNKNOWN'

    attempts = 1
    while True:
        try:
            output = func(*args, **kwargs)
        except AbortionError as e:
            # pylint doesn't like grpc's crazy exception hierarchy
            # pylint: disable=raising-non-exception
            if e.code not in RETRYABLE_ERRORS:
                raise e
            if attempts >= MAX_ATTEMPTS:
                print('Operation {} failed after {} attempts'.format(
                    operation, attempts
                ))
                raise e
        else:
            return output

        base_delay = min((BACKOFF_COEFF * (2 ** attempts)) / 1000.0, 60.0)
        delay = random.uniform(0, base_delay)
        print('Operation {} failed, retrying in {}s'.format(
            operation, delay))
        time.sleep(delay)
        attempts += 1



import cbor
from grpc.beta import implementations
from mediachain.rpc.utils import with_retry
from mediachain.datastore.data_objects import MultihashReference
from mediachain.proto import Transactor_pb2  # pylint: disable=no-name-in-module
from mediachain.proto import Types_pb2  # pylint: disable=no-name-in-module
from mediachain.transactor.blockchain_follower import BlockchainFollower
import mediachain.reader.api as reader


TIMEOUT_SECS = 120


class TransactorClient(object):

    def __init__(self, host, port):
        self.host = host
        self.port = port
        channel = implementations.insecure_channel(host, port)
        self.client = Transactor_pb2.beta_create_TransactorService_stub(channel)

    def insert_canonical(self, record, timeout=TIMEOUT_SECS):
        try:
            cbor_bytes = record.to_cbor_bytes()
        except AttributeError:
            cbor_bytes = cbor.dumps(record)

        req = Transactor_pb2.InsertRequest(canonicalCbor=cbor_bytes)
        ref = with_retry(self.client.InsertCanonical, req, timeout)
        return MultihashReference.from_base58(ref.reference)

    def get_chain_head(self, ref, timeout=TIMEOUT_SECS):
        req = Types_pb2.MultihashReference(reference=ref)
        response = with_retry(self.client.LookupChain, req, timeout)
        if response.WhichOneof('reference') == 'chain':
            return MultihashReference.from_base58(
                response.chain.reference
            )
        return None

    def update_chain(self, cell, timeout=TIMEOUT_SECS):
        try:
            cbor_bytes = cell.to_cbor_bytes()
        except AttributeError:
            cbor_bytes = cbor.dumps(cell)

        req = Transactor_pb2.UpdateRequest(chainCellCbor=cbor_bytes)
        ref = with_retry(self.client.UpdateChain, req, timeout)
        return MultihashReference.from_base58(ref.reference)

    def journal_stream(self, catchup=True, timeout=None):
        """
        A stream of journal events from the mediachain transactor network.
        If `catchup` is True (the default), will fetch the entire history
        of the blockchain and play back all historical events, followed by
        any incoming events.

        If `catchup` is False, only future events will be delivered.

        Defaults to an infinite stream with no timeout; you can force the
        stream to expire by passing in a `timeout` in seconds.

        IMPORTANT:
        You must close the stream by calling the `.cancel()` method on the
        returned `BlockchainFollower` object before exiting the process,
        otherwise the gRPC connection will keep the process alive indefinitely.

        To make cleanup automatic use with/as:
            with client.journal_stream() as stream:
              for event in stream:
                # handle event...


        :param catchup:
        :param timeout:
        :return:
        """
        req = Transactor_pb2.JournalStreamRequest()
        stream = self.client.JournalStream(req, timeout)
        follower = BlockchainFollower(stream, catchup)
        follower.start()
        return follower


    def canonical_stream(self, timeout=600):
        for event in self.journal_stream(timeout=timeout):
            if event.WhichOneof("event") == "updateChainEvent":
                ref = event.updateChainEvent.canonical.reference
                obj = reader.get_object(self, ref)
                yield obj




import sys
import os
import subprocess
import time
import argparse

def ingest_dirs(top):
    for (dirpath, dirnames, filenames) in os.walk(top):
        if len(dirnames) == 0 and len(filenames) > 0:
            yield dirpath

def ingest(ns):
    procs = dict()
    def spawn(adir):
        print "Ingest %s" % adir
        proc = subprocess.Popen([ns.script, adir])
        procs[proc.pid] = proc
    
    def reap_some():
        done = []
        while True:
            for (pid, proc) in procs.items():
                if proc.poll() is not None:
                    done.append(pid)
            
            if len(done) > 0:
                for pid in done:
                    del procs[pid]
                break
            else:
                time.sleep(1)
        
    for adir in ingest_dirs(ns.dir):
        if (len(procs) < ns.procs):
            spawn(adir)
        else:
            reap_some()
            spawn(adir)
    
    while len(procs) > 0:
        reap_some()

def main(args):
    parser = argparse.ArgumentParser(
        prog = "ingest-parallel.py",
        description = "run parallel ingestion scripts in the leaf directories of a dataset"
    )
    parser.add_argument('-n', '--processes',
                        type = int,
                        default = 10,
                        dest = 'procs',
                        help = "Number of parallel ingestion processes")
    parser.add_argument('script',
                        type = str,
                        help = "Ingest script; must accept a directory and be nice!")
    parser.add_argument('dir',
                        type = str,
                        help = "Top dataset directory")
    ns = parser.parse_args(args)
    ingest(ns)
                        
    
if __name__ == '__main__':
    main(sys.argv[1:])



"""Multihash implementation in Python."""

import hashlib
import struct
import sys

import six

# Optional SHA-3 hashing via pysha3
try:
    import sha3
except ImportError:
    sha3 = None

# Optional BLAKE2 hashing via pyblake2
try:
    import pyblake2
except ImportError:
    pyblake2 = None


# Constants
SHA1 = 0x11
SHA2_256 = 0x12
SHA2_512 = 0x13
SHA3 = 0x14
BLAKE2B = 0x40
BLAKE2S = 0x41

NAMES = {
    'sha1':     SHA1,
    'sha2-256': SHA2_256,
    'sha2-512': SHA2_512,
    'sha3':     SHA3,
    'blake2b':  BLAKE2B,
    'blake2s':  BLAKE2S,
}

CODES = dict((v, k) for k, v in NAMES.items())

LENGTHS = {
    'sha1':    20,
    'sha256':  32,
    'sha512':  64,
    'sha3':    64,
    'blake2b': 64,
    'blake2s': 32,
}

FUNCS = {
    SHA1: hashlib.sha1,
    SHA2_256: hashlib.sha256,
    SHA2_512: hashlib.sha512,
}

if sha3:
    FUNCS[SHA3] = lambda: hashlib.new('sha3_512')

if pyblake2:
    FUNCS[BLAKE2B] = lambda: pyblake2.blake2b()
    FUNCS[BLAKE2S] = lambda: pyblake2.blake2s()


def _hashfn(hashfn):
    """Return an initialised hash object, by function, name or integer id

    >>> _hashfn(SHA1) # doctest: +ELLIPSIS
    <sha1 HASH object @ 0x...>

    >>> _hashfn('sha2-256') # doctest: +ELLIPSIS
    <sha256 HASH object @ 0x...>
    >>> _hashfn('18') # doctest: +ELLIPSIS
    <sha256 HASH object @ 0x...>

    >>> _hashfn('md5')
    Traceback (most recent call last):
      ...
    ValueError: Unknown hash function "md5"
    """
    if six.callable(hashfn):
        return hashfn()

    elif isinstance(hashfn, six.integer_types):
        return FUNCS[hashfn]()

    elif isinstance(hashfn, six.string_types):
        if hashfn in NAMES:
            return FUNCS[NAMES[hashfn]]()

        elif hashfn.isdigit():
            return _hashfn(int(hashfn))

    raise ValueError('Unknown hash function "{0}"'.format(hashfn))


def is_app_code(code):
    """Check if the code is an application specific code.

    >>> is_app_code(SHA1)
    False
    >>> is_app_code(0)
    True
    """
    if isinstance(code, six.integer_types):
        return code >= 0 and code < 0x10

    else:
        return False


def is_valid_code(code):
    """Check if the digest algorithm code is valid.

    >>> is_valid_code(SHA1)
    True
    >>> is_valid_code(0)
    True
    """
    if is_app_code(code):
        return True

    elif isinstance(code, six.integer_types):
        return code in CODES

    else:
        return False


def decode(buf):
    r"""Decode a hash from the given Multihash.

    After validating the hash type and length in the two prefix bytes, this
    function removes them and returns the raw hash.

    >>> encoded = b'\x11\x14\xc3\xd4XGWbx`AAh\x01%\xa4o\xef9Nl('
    >>> bytearray(decode(encoded))
    bytearray(b'\xc3\xd4XGWbx`AAh\x01%\xa4o\xef9Nl(')

    >>> decode(encoded) == encoded[2:] == hashlib.sha1(b'thanked').digest()
    True
    """
    if len(buf) < 3:
        raise ValueError('Buffer too short')

    if len(buf) > 129:
        raise ValueError('Buffer too long')

    code, length = struct.unpack('BB', buf[:2])

    if not is_valid_code(code):
        raise ValueError('Invalid code "{0}"'.format(code))

    digest = buf[2:]
    if len(digest) != length:
        raise ValueError('Inconsistent length ({0} != {1})'.format(
            len(digest), length))

    return digest


def encode(content, code):
    """Encode a binary or text string using the digest function corresponding
    to the given code.  Returns the hash of the content, prefixed with the
    code and the length of the digest, according to the Multihash spec.

    >>> encoded = encode('testing', SHA1)
    >>> len(encoded)
    22
    >>> encoded[:2]
    bytearray(b'\\x11\\x07')

    >>> encoded = encode('works with sha3?', SHA3)
    >>> len(encoded)
    66
    >>> encoded[:2]
    bytearray(b'\\x14\\x10')
    """
    if not is_valid_code(code):
        raise TypeError('Unknown code')

    hashfn = _hashfn(code)

    if isinstance(content, six.binary_type):
        hashfn.update(content)
    elif isinstance(content, six.string_types):
        hashfn.update(content.encode('utf-8'))

    digest = hashfn.digest()
    if len(digest) > 127:
        raise ValueError('Multihash does not support digest length > 127')

    output = bytearray([code, len(digest)])
    output.extend(digest)
    return output




import ipfsApi
from requests.exceptions import ConnectionError
from tempfile import NamedTemporaryFile
from mediachain.datastore.utils import multihash_ref, object_for_bytes

__IPFS_CONFIG = {'host': 'localhost', 'port': 5001}


def set_ipfs_config(cfg):
    global __IPFS_CONFIG
    __IPFS_CONFIG = cfg


def get_ipfs_config():
    global __IPFS_CONFIG
    return __IPFS_CONFIG


def get_ipfs_datastore():
    global __IPFS_CONFIG
    return IpfsDatastore(**__IPFS_CONFIG)


class IpfsDatastore(object):
    def __init__(self, host='localhost', port=5001):
        self.client = ipfsApi.Client(host, port)
        try:
            self.client.id()
        except ConnectionError:
            raise Exception(
                'Unable to connect to ipfs API at {}:{} \n'
                'For ipfs installation instructions see '
                'https://ipfs.io/docs/install'.format(host, port)
            )

    def put(self, data_object):
        try:
            content = data_object.to_cbor_bytes()
        except AttributeError:
            content = data_object

        with NamedTemporaryFile() as f:
            f.write(content)
            f.flush()
            result = self.client.add(f.name)

        if isinstance(result, basestring):
            return result

        # when adding a file, we only want the entry pointing to the file's
        # contents, not the entries for the intermediate directories.
        # if the result is a list, the first entry will be, e.g:
        # {u'Bytes': 22, u'Name': u'/tmp/foo/bar/tmp7Hobfu'}
        # followed by entries for the file, plus entries for '/tmp', '/tmp/foo',
        # etc.
        header = result.pop(0)
        if 'Hash' in header:
            return header['Hash']

        name = header['Name']
        hashes = [h['Hash'] for h in result if h['Name'] == name]
        return multihash_ref(hashes[0])

    def get(self, ref):
        stream = self.client.cat(str(ref), stream=True)
        byte_string = stream.read()
        return object_for_bytes(byte_string)




# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: Transactor.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import Types_pb2 as Types__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='Transactor.proto',
  package='protocol',
  syntax='proto3',
  serialized_pb=_b('\n\x10Transactor.proto\x12\x08protocol\x1a\x0bTypes.proto\"&\n\rInsertRequest\x12\x15\n\rcanonicalCbor\x18\x01 \x01(\x0c\"&\n\rUpdateRequest\x12\x15\n\rchainCellCbor\x18\x01 \x01(\x0c\"\x16\n\x14JournalStreamRequest\"\xa6\x01\n\x11UpdateChainResult\x12/\n\tcanonical\x18\x01 \x01(\x0b\x32\x1c.protocol.MultihashReference\x12+\n\x05\x63hain\x18\x02 \x01(\x0b\x32\x1c.protocol.MultihashReference\x12\x33\n\rchainPrevious\x18\x03 \x01(\x0b\x32\x1c.protocol.MultihashReference\"\xc9\x01\n\x0cJournalEvent\x12<\n\x14insertCanonicalEvent\x18\x01 \x01(\x0b\x32\x1c.protocol.MultihashReferenceH\x00\x12\x37\n\x10updateChainEvent\x18\x02 \x01(\x0b\x32\x1b.protocol.UpdateChainResultH\x00\x12\x39\n\x11journalBlockEvent\x18\x03 \x01(\x0b\x32\x1c.protocol.MultihashReferenceH\x00\x42\x07\n\x05\x65vent2\xb5\x02\n\x11TransactorService\x12H\n\x0fInsertCanonical\x12\x17.protocol.InsertRequest\x1a\x1c.protocol.MultihashReference\x12\x44\n\x0bUpdateChain\x12\x17.protocol.UpdateRequest\x1a\x1c.protocol.MultihashReference\x12\x45\n\x0bLookupChain\x12\x1c.protocol.MultihashReference\x1a\x18.protocol.ChainReference\x12I\n\rJournalStream\x12\x1e.protocol.JournalStreamRequest\x1a\x16.protocol.JournalEvent0\x01\x42#\n!io.mediachain.protocol.transactorb\x06proto3')
  ,
  dependencies=[Types__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_INSERTREQUEST = _descriptor.Descriptor(
  name='InsertRequest',
  full_name='protocol.InsertRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='canonicalCbor', full_name='protocol.InsertRequest.canonicalCbor', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=43,
  serialized_end=81,
)


_UPDATEREQUEST = _descriptor.Descriptor(
  name='UpdateRequest',
  full_name='protocol.UpdateRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='chainCellCbor', full_name='protocol.UpdateRequest.chainCellCbor', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=83,
  serialized_end=121,
)


_JOURNALSTREAMREQUEST = _descriptor.Descriptor(
  name='JournalStreamRequest',
  full_name='protocol.JournalStreamRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=123,
  serialized_end=145,
)


_UPDATECHAINRESULT = _descriptor.Descriptor(
  name='UpdateChainResult',
  full_name='protocol.UpdateChainResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='canonical', full_name='protocol.UpdateChainResult.canonical', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chain', full_name='protocol.UpdateChainResult.chain', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chainPrevious', full_name='protocol.UpdateChainResult.chainPrevious', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=148,
  serialized_end=314,
)


_JOURNALEVENT = _descriptor.Descriptor(
  name='JournalEvent',
  full_name='protocol.JournalEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='insertCanonicalEvent', full_name='protocol.JournalEvent.insertCanonicalEvent', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='updateChainEvent', full_name='protocol.JournalEvent.updateChainEvent', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='journalBlockEvent', full_name='protocol.JournalEvent.journalBlockEvent', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='event', full_name='protocol.JournalEvent.event',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=317,
  serialized_end=518,
)

_UPDATECHAINRESULT.fields_by_name['canonical'].message_type = Types__pb2._MULTIHASHREFERENCE
_UPDATECHAINRESULT.fields_by_name['chain'].message_type = Types__pb2._MULTIHASHREFERENCE
_UPDATECHAINRESULT.fields_by_name['chainPrevious'].message_type = Types__pb2._MULTIHASHREFERENCE
_JOURNALEVENT.fields_by_name['insertCanonicalEvent'].message_type = Types__pb2._MULTIHASHREFERENCE
_JOURNALEVENT.fields_by_name['updateChainEvent'].message_type = _UPDATECHAINRESULT
_JOURNALEVENT.fields_by_name['journalBlockEvent'].message_type = Types__pb2._MULTIHASHREFERENCE
_JOURNALEVENT.oneofs_by_name['event'].fields.append(
  _JOURNALEVENT.fields_by_name['insertCanonicalEvent'])
_JOURNALEVENT.fields_by_name['insertCanonicalEvent'].containing_oneof = _JOURNALEVENT.oneofs_by_name['event']
_JOURNALEVENT.oneofs_by_name['event'].fields.append(
  _JOURNALEVENT.fields_by_name['updateChainEvent'])
_JOURNALEVENT.fields_by_name['updateChainEvent'].containing_oneof = _JOURNALEVENT.oneofs_by_name['event']
_JOURNALEVENT.oneofs_by_name['event'].fields.append(
  _JOURNALEVENT.fields_by_name['journalBlockEvent'])
_JOURNALEVENT.fields_by_name['journalBlockEvent'].containing_oneof = _JOURNALEVENT.oneofs_by_name['event']
DESCRIPTOR.message_types_by_name['InsertRequest'] = _INSERTREQUEST
DESCRIPTOR.message_types_by_name['UpdateRequest'] = _UPDATEREQUEST
DESCRIPTOR.message_types_by_name['JournalStreamRequest'] = _JOURNALSTREAMREQUEST
DESCRIPTOR.message_types_by_name['UpdateChainResult'] = _UPDATECHAINRESULT
DESCRIPTOR.message_types_by_name['JournalEvent'] = _JOURNALEVENT

InsertRequest = _reflection.GeneratedProtocolMessageType('InsertRequest', (_message.Message,), dict(
  DESCRIPTOR = _INSERTREQUEST,
  __module__ = 'Transactor_pb2'
  # @@protoc_insertion_point(class_scope:protocol.InsertRequest)
  ))
_sym_db.RegisterMessage(InsertRequest)

UpdateRequest = _reflection.GeneratedProtocolMessageType('UpdateRequest', (_message.Message,), dict(
  DESCRIPTOR = _UPDATEREQUEST,
  __module__ = 'Transactor_pb2'
  # @@protoc_insertion_point(class_scope:protocol.UpdateRequest)
  ))
_sym_db.RegisterMessage(UpdateRequest)

JournalStreamRequest = _reflection.GeneratedProtocolMessageType('JournalStreamRequest', (_message.Message,), dict(
  DESCRIPTOR = _JOURNALSTREAMREQUEST,
  __module__ = 'Transactor_pb2'
  # @@protoc_insertion_point(class_scope:protocol.JournalStreamRequest)
  ))
_sym_db.RegisterMessage(JournalStreamRequest)

UpdateChainResult = _reflection.GeneratedProtocolMessageType('UpdateChainResult', (_message.Message,), dict(
  DESCRIPTOR = _UPDATECHAINRESULT,
  __module__ = 'Transactor_pb2'
  # @@protoc_insertion_point(class_scope:protocol.UpdateChainResult)
  ))
_sym_db.RegisterMessage(UpdateChainResult)

JournalEvent = _reflection.GeneratedProtocolMessageType('JournalEvent', (_message.Message,), dict(
  DESCRIPTOR = _JOURNALEVENT,
  __module__ = 'Transactor_pb2'
  # @@protoc_insertion_point(class_scope:protocol.JournalEvent)
  ))
_sym_db.RegisterMessage(JournalEvent)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n!io.mediachain.protocol.transactor'))
import grpc
from grpc.beta import implementations as beta_implementations
from grpc.beta import interfaces as beta_interfaces
from grpc.framework.common import cardinality
from grpc.framework.interfaces.face import utilities as face_utilities


class TransactorServiceStub(object):

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.InsertCanonical = channel.unary_unary(
        '/protocol.TransactorService/InsertCanonical',
        request_serializer=InsertRequest.SerializeToString,
        response_deserializer=Types__pb2.MultihashReference.FromString,
        )
    self.UpdateChain = channel.unary_unary(
        '/protocol.TransactorService/UpdateChain',
        request_serializer=UpdateRequest.SerializeToString,
        response_deserializer=Types__pb2.MultihashReference.FromString,
        )
    self.LookupChain = channel.unary_unary(
        '/protocol.TransactorService/LookupChain',
        request_serializer=Types__pb2.MultihashReference.SerializeToString,
        response_deserializer=Types__pb2.ChainReference.FromString,
        )
    self.JournalStream = channel.unary_stream(
        '/protocol.TransactorService/JournalStream',
        request_serializer=JournalStreamRequest.SerializeToString,
        response_deserializer=JournalEvent.FromString,
        )


class TransactorServiceServicer(object):

  def InsertCanonical(self, request, context):
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UpdateChain(self, request, context):
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def LookupChain(self, request, context):
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def JournalStream(self, request, context):
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_TransactorServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'InsertCanonical': grpc.unary_unary_rpc_method_handler(
          servicer.InsertCanonical,
          request_deserializer=InsertRequest.FromString,
          response_serializer=Types__pb2.MultihashReference.SerializeToString,
      ),
      'UpdateChain': grpc.unary_unary_rpc_method_handler(
          servicer.UpdateChain,
          request_deserializer=UpdateRequest.FromString,
          response_serializer=Types__pb2.MultihashReference.SerializeToString,
      ),
      'LookupChain': grpc.unary_unary_rpc_method_handler(
          servicer.LookupChain,
          request_deserializer=Types__pb2.MultihashReference.FromString,
          response_serializer=Types__pb2.ChainReference.SerializeToString,
      ),
      'JournalStream': grpc.unary_stream_rpc_method_handler(
          servicer.JournalStream,
          request_deserializer=JournalStreamRequest.FromString,
          response_serializer=JournalEvent.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'protocol.TransactorService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class BetaTransactorServiceServicer(object):
  def InsertCanonical(self, request, context):
    context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
  def UpdateChain(self, request, context):
    context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
  def LookupChain(self, request, context):
    context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
  def JournalStream(self, request, context):
    context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)


class BetaTransactorServiceStub(object):
  def InsertCanonical(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
    raise NotImplementedError()
  InsertCanonical.future = None
  def UpdateChain(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
    raise NotImplementedError()
  UpdateChain.future = None
  def LookupChain(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
    raise NotImplementedError()
  LookupChain.future = None
  def JournalStream(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
    raise NotImplementedError()


def beta_create_TransactorService_server(servicer, pool=None, pool_size=None, default_timeout=None, maximum_timeout=None):
  request_deserializers = {
    ('protocol.TransactorService', 'InsertCanonical'): InsertRequest.FromString,
    ('protocol.TransactorService', 'JournalStream'): JournalStreamRequest.FromString,
    ('protocol.TransactorService', 'LookupChain'): Types__pb2.MultihashReference.FromString,
    ('protocol.TransactorService', 'UpdateChain'): UpdateRequest.FromString,
  }
  response_serializers = {
    ('protocol.TransactorService', 'InsertCanonical'): Types__pb2.MultihashReference.SerializeToString,
    ('protocol.TransactorService', 'JournalStream'): JournalEvent.SerializeToString,
    ('protocol.TransactorService', 'LookupChain'): Types__pb2.ChainReference.SerializeToString,
    ('protocol.TransactorService', 'UpdateChain'): Types__pb2.MultihashReference.SerializeToString,
  }
  method_implementations = {
    ('protocol.TransactorService', 'InsertCanonical'): face_utilities.unary_unary_inline(servicer.InsertCanonical),
    ('protocol.TransactorService', 'JournalStream'): face_utilities.unary_stream_inline(servicer.JournalStream),
    ('protocol.TransactorService', 'LookupChain'): face_utilities.unary_unary_inline(servicer.LookupChain),
    ('protocol.TransactorService', 'UpdateChain'): face_utilities.unary_unary_inline(servicer.UpdateChain),
  }
  server_options = beta_implementations.server_options(request_deserializers=request_deserializers, response_serializers=response_serializers, thread_pool=pool, thread_pool_size=pool_size, default_timeout=default_timeout, maximum_timeout=maximum_timeout)
  return beta_implementations.server(method_implementations, options=server_options)


def beta_create_TransactorService_stub(channel, host=None, metadata_transformer=None, pool=None, pool_size=None):
  request_serializers = {
    ('protocol.TransactorService', 'InsertCanonical'): InsertRequest.SerializeToString,
    ('protocol.TransactorService', 'JournalStream'): JournalStreamRequest.SerializeToString,
    ('protocol.TransactorService', 'LookupChain'): Types__pb2.MultihashReference.SerializeToString,
    ('protocol.TransactorService', 'UpdateChain'): UpdateRequest.SerializeToString,
  }
  response_deserializers = {
    ('protocol.TransactorService', 'InsertCanonical'): Types__pb2.MultihashReference.FromString,
    ('protocol.TransactorService', 'JournalStream'): JournalEvent.FromString,
    ('protocol.TransactorService', 'LookupChain'): Types__pb2.ChainReference.FromString,
    ('protocol.TransactorService', 'UpdateChain'): Types__pb2.MultihashReference.FromString,
  }
  cardinalities = {
    'InsertCanonical': cardinality.Cardinality.UNARY_UNARY,
    'JournalStream': cardinality.Cardinality.UNARY_STREAM,
    'LookupChain': cardinality.Cardinality.UNARY_UNARY,
    'UpdateChain': cardinality.Cardinality.UNARY_UNARY,
  }
  stub_options = beta_implementations.stub_options(host=host, metadata_transformer=metadata_transformer, request_serializers=request_serializers, response_deserializers=response_deserializers, thread_pool=pool, thread_pool_size=pool_size)
  return beta_implementations.dynamic_stub(channel, 'protocol.TransactorService', cardinalities, options=stub_options)
# @@protoc_insertion_point(module_scope)



import os
from urlparse import urlparse
import requests
import base64
from PIL import Image
from StringIO import StringIO


def load_asset(asset):
    try:
        uri = asset['uri']
    except (TypeError, KeyError):
        return None

    if uri.lower().startswith('file://'):
        p = urlparse(uri)
        file_path = os.path.abspath(os.path.join(p.netloc, p.path))
        if not os.path.isfile(file_path):
            return None
        try:
            with open(file_path) as f:
                return f.read()
        except IOError as e:
            print('error reading from {}: {}'.format(file_path, e))
            return

    try:
        r = requests.get(uri)
        return r.content
    except requests.exceptions.RequestException as e:
        print('error downloading media from {}: {}'.format(uri, e))

    return None


def process_asset(name, asset_data):
    if name == 'thumbnail':
        return resize_image(asset_data, THUMBNAIL_SIZE)
    return asset_data


# size constants for images
THUMBNAIL_SIZE = (150, 150)


def resize_image(image_data, size=THUMBNAIL_SIZE):
    img = Image.open(StringIO(image_data))
    if (img.size[0] > size[0]) or (img.size[1] > size[1]):
        img.thumbnail(size, Image.ANTIALIAS)
    buf = StringIO()
    img.save(buf, "JPEG")
    buf.seek(0)
    data = buf.read()
    return data


def make_jpeg_data_uri(data):
    return 'data:image/jpeg;base64,' + base64.urlsafe_b64encode(data)



import base58
import cbor
from mediachain.proto import Types_pb2  # pylint: disable=no-name-in-module
from mediachain.datastore.data_objects import MultihashReference


def multihash_ref(ref):
    if isinstance(ref, MultihashReference):
        return ref

    try:
        ref = ref.reference
    except AttributeError:
        pass
    try:
        ref = ref['@link']
    except (TypeError, KeyError):
        pass

    try:
        ref = base58.b58decode(ref)
    except ValueError:
        pass
    return MultihashReference(multihash=ref)


def rpc_ref(ref):
    return Types_pb2.MultihashReference(reference=ref_base58(ref))


def ref_base58(ref):
    ref_str = ref
    try:
        ref_str = ref.multihash_base58()
    except AttributeError:
        pass
    try:
        ref_str = ref.reference
    except AttributeError:
        pass
    try:
        ref_str = base58.b58encode(ref['@link'])
    except (KeyError, ValueError, TypeError):
        pass
    return ref_str


def bytes_for_object(obj):
    try:
        return obj.to_cbor_bytes()
    except AttributeError:
        pass

    if isinstance(obj, dict):
        try:
            return cbor.dumps(obj)
        except ValueError:
            pass
    return obj


def object_for_bytes(obj_bytes):
    try:
        return cbor.loads(obj_bytes)
    except ValueError:
        return obj_bytes



MEDIACHAIN_OBJECT_TAG = '__mediachain_object__'
MEDIACHAIN_ASSET_TAG = '__mediachain_asset__'


def is_tagged_dict(tag, value):
    if not isinstance(value, dict):
        return False
    return tag in value


def is_asset(value):
    return is_tagged_dict(MEDIACHAIN_ASSET_TAG, value)


def is_mediachain_object(value):
    return is_tagged_dict(MEDIACHAIN_OBJECT_TAG, value)


def is_canonical(obj):
    return obj['type'] == 'artefact' or obj['type'] == 'entity'




from ipfs import IpfsDatastore
from rpc import get_db

__USE_IPFS_FOR_RAW_DATA = True


def set_use_ipfs_for_raw_data(use):
    global __USE_IPFS_FOR_RAW_DATA
    __USE_IPFS_FOR_RAW_DATA = use


def get_raw_datastore():
    global __USE_IPFS_FOR_RAW_DATA
    if __USE_IPFS_FOR_RAW_DATA:
        return IpfsDatastore()

    return get_db()



from __future__ import unicode_literals
from __future__ import print_function

from copy import deepcopy
import re
from mediachain.ingestion.asset_loader import load_asset, process_asset
from mediachain.translation.utils import is_asset, is_canonical, \
    is_mediachain_object
from mediachain.datastore import get_db, get_raw_datastore
from grpc.framework.interfaces.face.face import AbortionError
from mediachain.datastore.utils import multihash_ref
import sys
import traceback


def print_err(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


class Writer(object):
    def __init__(self, transactor, datastore=None,
                 download_remote_assets=False):
        self.transactor = transactor
        self.datastore = datastore or get_db()
        self.download_remote_assets = download_remote_assets

    def write_dataset(self, dataset_iterator):
        translator_id = dataset_iterator.translator.translator_id()

        for result in dataset_iterator:
            translated = result['translated']
            raw = result['raw_content']
            local_assets = result.get('local_assets', {})
            try:
                refs = self.submit_translator_output(translator_id, translated,
                                                     raw, local_assets)
                yield refs
            except AbortionError:
                for line in traceback.format_exception(*sys.exc_info()):
                    print_err(line.rstrip('\n'))

    def submit_translator_output(self,
                                 translator_id,
                                 translated,
                                 raw_content,
                                 local_assets):
        meta_source = multihash_ref(self.datastore.put(raw_content))

        raw_ref = self.store_raw(raw_content)

        common_meta = {'translator': translator_id}
        canonical_meta = common_meta.copy()
        canonical_meta.update({'raw_ref': raw_ref.to_map()})

        canonical = translated['canonical']
        canonical = self.flatten_record(canonical, canonical_meta,
                                        local_assets, meta_source=meta_source)
        canonical_ref = self.submit_object(canonical)

        chain = translated.get('chain', [])
        chain_refs = []
        for cell in chain:
            cell = self.flatten_record(cell, common_meta, local_assets)
            cell['ref'] = canonical_ref.to_map()
            chain_ref = self.submit_object(cell)
            chain_refs += chain_ref.multihash_base58()
        return {
            'canonical': canonical_ref.multihash_base58(),
            'chain': chain_refs
        }

    def flatten_record(self, record, common_meta, local_assets,
                       meta_source=None):
        record = deepcopy(record)
        assets = {k: v for k, v in record.iteritems()
                  if is_asset(v)}

        for k, a in assets.iteritems():
            local = local_assets.get(k, None)
            ref = self.store_asset(k, local, a)
            if ref is None:
                # print('Unable to store asset with key {}, removing'.format(k))
                del record[k]
            else:
                record[k] = ref.to_map()

        objects = {k: v for k, v in record.iteritems()
                   if isinstance(v, dict)}

        for k, o in objects.iteritems():
            flat = self.flatten_record(o, common_meta,
                                       meta_source, local_assets)
            if is_mediachain_object(o):
                ref = self.submit_object(flat)
                record[k] = ref.to_map()
            else:
                record[k] = flat

        if is_mediachain_object(record):
            del record['__mediachain_object__']
            try:
                record['metaSource'] = meta_source.to_map()
            except (TypeError, AttributeError):
                pass
            record['meta'].update(common_meta)
        return record

    def store_raw(self, raw_content):
        store = get_raw_datastore()
        ref = store.put(raw_content)
        return multihash_ref(ref)

    def store_asset(self, name, local_asset, remote_asset):
        data = load_asset(local_asset)
        if data is None and self.download_remote_assets:
            data = load_asset(remote_asset)
        if data is None:
            return None
        data = process_asset(name, data)
        return self.store_raw(data)

    def submit_object(self, obj):
        try:
            if is_canonical(obj):
                return self.transactor.insert_canonical(obj)
            else:
                return self.transactor.update_chain(obj)
        except AbortionError as e:
            ref_str = duplicate_canonical_ref(e.details)
            if ref_str is None:
                raise

        return multihash_ref(ref_str)


def merge_meta(obj, meta):
    merged = {'data': {}}
    merged.update(meta)
    if 'meta' in obj:
        merged.update(obj['meta'])
    obj['meta'] = merged


# TODO(yusef): replace string parsing hack with grpc error metadata
# requires support in grpc-java (in master but not yet released)

# this regex will only work for base58 encoded sha2-256 hashes,
# and will break if we ever change the error message :(
duplicate_insert_pattern = re.compile(
    '.*Duplicate Journal Insert.*(Qm[1-9a-km-zA-HJ-NP-Z]{44}).*'
)


def duplicate_canonical_ref(grpc_error_str):
    m = re.match(duplicate_insert_pattern, grpc_error_str)
    if not m:
        return None
    return m.group(1)





import sys
import argparse
import os
import traceback
from time import sleep

# Set the GRPC log level to ERROR, unless it's already been set
# This needs to happen before importing anything that will
# import grpc, since it reads the environment var immediately on import
def set_grpc_verbosity():
    level = os.environ.get('GRPC_VERBOSITY', 'ERROR')
    os.environ['GRPC_VERBOSITY'] = level

set_grpc_verbosity()

from mediachain.reader import api
from mediachain.datastore import set_use_ipfs_for_raw_data
from mediachain.datastore.ipfs import set_ipfs_config
from mediachain.datastore.rpc import set_rpc_datastore_config, close_db
from mediachain.writer import Writer
from mediachain.translation import get_translator
from mediachain.ingestion.directory_iterator import LocalFileIterator
from mediachain.transactor.client import TransactorClient


def main(arguments=None):
    def configure_datastore(ns):
        host = getattr(ns, 'datastore_host')
        if host is None:
            host = getattr(ns, 'host')
        port = getattr(ns, 'datastore_port')
        set_rpc_datastore_config({'host': host, 'port': port})

    def configure_ipfs(ns):
        try:
            cfg = {'host': getattr(ns, 'ipfs_host'),
                   'port': getattr(ns, 'ipfs_port')
                   }

            set_ipfs_config(cfg)

            if getattr(ns, 'disable_ipfs'):
                set_use_ipfs_for_raw_data(False)
        except AttributeError:
            pass

    if arguments == None:
        arguments = sys.argv[1:]

    parser = argparse.ArgumentParser(
        prog='mediachain',
        description='Mediachain CLI'
    )

    parser.add_argument('-s', '--host',
                        type=str,
                        default='facade.mediachain.io',
                        dest='host')
    parser.add_argument('-p', '--port',
                        type=int,
                        default=10001,
                        dest='port')
    parser.add_argument('--datastore-host',
                        dest='datastore_host',
                        type=str,
                        required=False,
                        help='Hostname or ip address of datastore service. ' +
                        'If not given, will default to the value of --host')
    parser.add_argument('--datastore-port',
                        dest='datastore_port',
                        type=int,
                        default=10002,
                        help='Port to use when connecting to the datastore ' +
                             'service.')
    parser.add_argument('--disable-ipfs',
                        dest='disable_ipfs',
                        action='store_true',
                        help=('If set, do not upload images and raw metadata '
                              'to IPFS')
                        )
    parser.add_argument('--ipfs-host',
                        dest='ipfs_host',
                        type=str,
                        default='localhost',
                        help='Hostname or ip address of IPFS api server'
                        )
    parser.add_argument('--ipfs-port',
                        dest='ipfs_port',
                        type=int,
                        default=5001,
                        help='Port for IPFS api server')

    subparsers = parser.add_subparsers(help='Mediachain Reader SubCommands',
                                       dest='subcommand')

    get_parser = subparsers.add_parser(
        'get',
        help='Get a revision chain for a given artefact/entity id'
    )
    get_parser.add_argument('object_id',
                            type=str,
                            help='The id of the artefact/entity to fetch')

    ingest_parser = subparsers.add_parser(
        'ingest',
        help='Ingest a directory of scraped Getty JSON data'
    )
    ingest_parser.add_argument('translator_id',
                               type=str,
                               help='identifier for a schema translator' +
                               'e.g. GettyTranslator/0.1')
    ingest_parser.add_argument('dir',
                               type=str,
                               help='Path to getty json directory root')
    ingest_parser.add_argument('--skip-validation',
                               dest='skip_validation',
                               action='store_true',
                               help=argparse.SUPPRESS)  # hide from --help
    ingest_parser.add_argument('-m', '--max-entries',
                               type=int,
                               dest='max_num',
                               help='Max json entries to parse. ' +
                               'Defaults to 0 (no maximum)',
                               default=0)
    ingest_parser.add_argument('-d', '--download-thumbnails',
                               type=bool,
                               dest='download_thumbs',
                               help='If set, download thumbnails if not found' +
                                    ' on disk.',
                               default=False)


    def get_cmd(ns):
        transactor = TransactorClient(ns.host, ns.port)
        api.get_and_print_object(transactor, ns.object_id)

    def ingest_cmd(args):
        translator = get_translator(args.translator_id)
        iterator = LocalFileIterator(translator, args.dir, args.max_num)

        transactor = TransactorClient(args.host, args.port)
        writer = Writer(transactor, download_remote_assets=args.download_thumbs)

        for refs in writer.write_dataset(iterator):
            print('Inserted canonical: {}'.format(refs['canonical']))

    SUBCOMMANDS={
        'get': get_cmd,
        'ingest': ingest_cmd
    }

    ns = parser.parse_args(arguments)
    if getattr(ns, 'skip_validation', False):
        os.environ['MEDIACHAIN_SKIP_SCHEMA_VALIDATION'] = 'true'

    fn = SUBCOMMANDS[ns.subcommand]

    configure_datastore(ns)
    configure_ipfs(ns)

    try:
        fn(ns)
        close_db()
    except KeyboardInterrupt:
        for line in traceback.format_exception(*sys.exc_info()):
            print line,
        sleep(1)
        os._exit(-1)

if __name__ == "__main__":
    main()



import json
import os
from jsonschema import validate as validate_schema
from mediachain.translation.utils import is_mediachain_object, is_canonical


class Translator(object):
    @staticmethod
    def translator_id():
        """
        :return: A unique string identifier for the translator
        """
        raise NotImplementedError("subclasses should return a unique string id")

    @staticmethod
    def parse(raw_metadata_bytes):
        """
        Parse a bytestring containing the raw metadata into a dict that's
         usable by the `translate` method.

         The default implementation is to parse the input as a uft-8 formatted
         json string.  Subclasses can override if they need to accept other
         formats.

        :param raw_metadata_bytes: a bytestring containing the raw metadata
        :return: a dict representation of the raw metadata
        """
        try:
            return json.loads(raw_metadata_bytes, 'utf-8')
        except (TypeError, ValueError):
            raise NotImplementedError(
                "subclasses should parse raw metadata if they accept formats" +
                " other than utf-8 json text"
            )

    @staticmethod
    def _translate(parsed_metadata):
        """
        Transforms the parsed metadata into a format suitable for writing to
        the mediachain network.
        :param parsed_metadata: a dict containing un-translated metadata
        :return: a dict containing mediachain records.


        e.g:
        {"object": {"type": "artefact", "meta": {}},
         "related": {
           "artefactCreatedBy": {
             "object": {"type": "entity", "meta": {}}
           }
         }
        }
        """
        raise NotImplementedError(
            "subclasses should return dictionary of translated metadata"
        )

    @classmethod
    def translate(cls, parsed_metadata):
        """
        Public wrapper for _translate. Don't override this
        """
        res = cls._translate(parsed_metadata)
        if not os.environ.get('MEDIACHAIN_SKIP_SCHEMA_VALIDATION', False):
            validated = cls.validate(res)
            print "Found {} valid Mediachain cells".format(validated)

        return res

    @staticmethod
    def can_translate_file(file_path):
        """
        Called by the ingestor to determine if this translator can accept
        a given local file.
        :param file_path: path to local file
        :return: bool indicating whether translation is possible
        """

        return False

    @classmethod
    def get_schema(self):
        # TODO: set BASE_DIR somewhere sane
        schema_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), 'schema.json'))
        if not hasattr(self, '_schema'):
            with open(schema_path) as schema_file:
                self._schema = json.load(schema_file)
        return self._schema

    @classmethod
    def validate(self, obj):
        if type(obj) in (int, str, bool, unicode):
            return 0

        validated = 0

        # TODO: (or is_update...)
        if is_mediachain_object(obj) and is_canonical(obj):
            validate_schema(obj['meta'], self.get_schema())
            validated = 1

        objects = {k: v for k, v in obj.iteritems()
                    if isinstance(v, dict)}

        lists = {k: v for k,v in obj.iteritems()
                    if isinstance(v, list)}
        objects_from_lists = [item for sublist in lists.values() for item in sublist]

        for o in objects.values() + objects_from_lists:
            validated = validated + self.validate(o)

        return validated






class DatasetIterator(object):
    def __init__(self, translator, limit=None):
        self.translator = translator
        self.limit = limit
        self._iterator = None

    def __iter__(self):
        return self

    def next(self):
        if self._iterator is None:
            self._iterator = self.generator()
        return next(self._iterator)

    def generator(self):
        raise NotImplementedError(
            'subclasses should yield translated metadata and supporting data'
        )



import threading
import time
from base58 import b58encode
from Queue import Queue, Empty as QueueEmpty
from collections import deque
from mediachain.transactor.block_cache import get_block_cache
from mediachain.proto import Transactor_pb2  # pylint: disable=no-name-in-module
from mediachain.datastore.utils import ref_base58
from grpc.beta.interfaces import StatusCode
from grpc.framework.interfaces.face.face import AbortionError

class BlockchainFollower(object):
    def __init__(self,
                 journal_stream,
                 catchup = True,
                 block_cache = None):
        if block_cache is None:
            block_cache = get_block_cache()

        self.cache = block_cache
        self.journal_stream = journal_stream
        self.block_ref_queue = Queue()
        self.incoming_event_queue = Queue()
        self.caught_up = False
        self.should_catchup = catchup
        self.first_incoming_event_received = False
        self.catchup_thread = threading.Thread(
            name='blockchain-catchup',
            target=self._perform_catchup)
        self.incoming_event_thread = threading.Thread(
            name='journal-stream-listener',
            target=self._receive_incoming_events
        )
        self.replay_stack = deque()
        self._event_iterator = self._event_stream()
        self._cancelled = False

    def __iter__(self):
        return self

    def next(self):
        return next(self._event_iterator)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cancel()
        return False

    def start(self):
        self.catchup_thread.start()
        self.incoming_event_thread.start()

    def cancel(self):
        # print('BlockchainFollower cancel')
        self._cancelled = True
        self.block_ref_queue.put('__abort__')
        self.incoming_event_queue.put('__abort__')
        self.journal_stream.cancel()

    def _perform_catchup(self):
        if not self.should_catchup:
            return

        while True:
            ref = self.block_ref_queue.get()
            if ref == '__abort__':
                return

            block = self.cache.get(ref)

            if block is None:
                print('Could not get block with ref {}'.format(ref))
                return

            self.replay_stack.appendleft(ref)

            chain = chain_ref(block)
            if chain is None:
                # print('Reached genesis block {}'.format(ref))
                self.caught_up = True
                return
            self.block_ref_queue.put(chain)

    def _receive_incoming_events(self):
        try:
            for event in self.journal_stream:
                if not self.first_incoming_event_received:
                    self.first_incoming_event_received = True
                    block_ref = block_event_ref(event)
                    if block_ref is None:
                        self.block_ref_queue.put('__abort__')
                        self.caught_up = True
                    else:
                        self.block_ref_queue.put(block_ref)
                self.incoming_event_queue.put(event)
        except AbortionError as e:
            if self._cancelled and e.code == StatusCode.CANCELLED:
                return
            else:
                raise


    def _event_stream(self):
        # block until catchup thread completes
        self.catchup_thread.join()

        for block_ref in self.replay_stack:
            block = self.cache.get(block_ref)
            entries = block.get('entries', [])
            for e in entries:
                yield block_event_to_rpc_event(e)
            block_event = Transactor_pb2.JournalEvent()
            block_event.journalBlockEvent.reference = ref_base58(block_ref)
            yield block_event

        while True:
            try:
                e = self.incoming_event_queue.get(block=False, timeout=1)
                if e == '__abort__':
                    return
                yield e
            except QueueEmpty:
                time.sleep(0.2)


def chain_ref(block):
    try:
        ref_bytes = bytes(block['chain']['@link'])
        return b58encode(ref_bytes)
    except (KeyError, ValueError):
        return None


def block_event_ref(rpc_event):
    if rpc_event.WhichOneof('event') != 'journalBlockEvent':
        return None
    return rpc_event.journalBlockEvent.reference


def block_event_to_rpc_event(event):
    rpc_event = Transactor_pb2.JournalEvent()
    if event['type'] == 'insert':
        rpc_event.insertCanonicalEvent.reference = ref_base58(event['ref'])
    elif event['type'] == 'update':
        rpc_event.updateChainEvent.canonical.reference = ref_base58(
            event['ref'])
        rpc_event.updateChainEvent.chain.reference = ref_base58(event['chain'])
        if 'chainPrevious' in event:
            rpc_event.updateChainEvent.chainPrevious.reference = ref_base58(
                event['chainPrevious'])
    else:
        raise ValueError('unknown journal event type: {}'.format(
            event['type']
        ))
    return rpc_event


# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: Datastore.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import Types_pb2 as Types__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='Datastore.proto',
  package='protocol',
  syntax='proto3',
  serialized_pb=_b('\n\x0f\x44\x61tastore.proto\x12\x08protocol\x1a\x0bTypes.proto\"\x1a\n\nDataObject\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x32\x88\x01\n\x10\x44\x61tastoreService\x12\x39\n\x03put\x12\x14.protocol.DataObject\x1a\x1c.protocol.MultihashReference\x12\x39\n\x03get\x12\x1c.protocol.MultihashReference\x1a\x14.protocol.DataObjectB\"\n io.mediachain.protocol.datastoreb\x06proto3')
  ,
  dependencies=[Types__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_DATAOBJECT = _descriptor.Descriptor(
  name='DataObject',
  full_name='protocol.DataObject',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='protocol.DataObject.data', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=42,
  serialized_end=68,
)

DESCRIPTOR.message_types_by_name['DataObject'] = _DATAOBJECT

DataObject = _reflection.GeneratedProtocolMessageType('DataObject', (_message.Message,), dict(
  DESCRIPTOR = _DATAOBJECT,
  __module__ = 'Datastore_pb2'
  # @@protoc_insertion_point(class_scope:protocol.DataObject)
  ))
_sym_db.RegisterMessage(DataObject)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\n io.mediachain.protocol.datastore'))
import grpc
from grpc.beta import implementations as beta_implementations
from grpc.beta import interfaces as beta_interfaces
from grpc.framework.common import cardinality
from grpc.framework.interfaces.face import utilities as face_utilities


class DatastoreServiceStub(object):

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.put = channel.unary_unary(
        '/protocol.DatastoreService/put',
        request_serializer=DataObject.SerializeToString,
        response_deserializer=Types__pb2.MultihashReference.FromString,
        )
    self.get = channel.unary_unary(
        '/protocol.DatastoreService/get',
        request_serializer=Types__pb2.MultihashReference.SerializeToString,
        response_deserializer=DataObject.FromString,
        )


class DatastoreServiceServicer(object):

  def put(self, request, context):
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def get(self, request, context):
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_DatastoreServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'put': grpc.unary_unary_rpc_method_handler(
          servicer.put,
          request_deserializer=DataObject.FromString,
          response_serializer=Types__pb2.MultihashReference.SerializeToString,
      ),
      'get': grpc.unary_unary_rpc_method_handler(
          servicer.get,
          request_deserializer=Types__pb2.MultihashReference.FromString,
          response_serializer=DataObject.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'protocol.DatastoreService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class BetaDatastoreServiceServicer(object):
  def put(self, request, context):
    context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
  def get(self, request, context):
    context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)


class BetaDatastoreServiceStub(object):
  def put(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
    raise NotImplementedError()
  put.future = None
  def get(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
    raise NotImplementedError()
  get.future = None


def beta_create_DatastoreService_server(servicer, pool=None, pool_size=None, default_timeout=None, maximum_timeout=None):
  request_deserializers = {
    ('protocol.DatastoreService', 'get'): Types__pb2.MultihashReference.FromString,
    ('protocol.DatastoreService', 'put'): DataObject.FromString,
  }
  response_serializers = {
    ('protocol.DatastoreService', 'get'): DataObject.SerializeToString,
    ('protocol.DatastoreService', 'put'): Types__pb2.MultihashReference.SerializeToString,
  }
  method_implementations = {
    ('protocol.DatastoreService', 'get'): face_utilities.unary_unary_inline(servicer.get),
    ('protocol.DatastoreService', 'put'): face_utilities.unary_unary_inline(servicer.put),
  }
  server_options = beta_implementations.server_options(request_deserializers=request_deserializers, response_serializers=response_serializers, thread_pool=pool, thread_pool_size=pool_size, default_timeout=default_timeout, maximum_timeout=maximum_timeout)
  return beta_implementations.server(method_implementations, options=server_options)


def beta_create_DatastoreService_stub(channel, host=None, metadata_transformer=None, pool=None, pool_size=None):
  request_serializers = {
    ('protocol.DatastoreService', 'get'): Types__pb2.MultihashReference.SerializeToString,
    ('protocol.DatastoreService', 'put'): DataObject.SerializeToString,
  }
  response_deserializers = {
    ('protocol.DatastoreService', 'get'): DataObject.FromString,
    ('protocol.DatastoreService', 'put'): Types__pb2.MultihashReference.FromString,
  }
  cardinalities = {
    'get': cardinality.Cardinality.UNARY_UNARY,
    'put': cardinality.Cardinality.UNARY_UNARY,
  }
  stub_options = beta_implementations.stub_options(host=host, metadata_transformer=metadata_transformer, request_serializers=request_serializers, response_deserializers=response_deserializers, thread_pool=pool, thread_pool_size=pool_size)
  return beta_implementations.dynamic_stub(channel, 'protocol.DatastoreService', cardinalities, options=stub_options)
# @@protoc_insertion_point(module_scope)



import cbor
from multihash import SHA2_256, encode as multihash_encode
from base58 import b58encode, b58decode


class MultihashReference(object):
    multihash = None

    @classmethod
    def for_content(cls, content_bytes):
        content_hash = multihash_encode(bytes(content_bytes), SHA2_256)
        return cls(content_hash)

    @classmethod
    def from_base58(cls, base58str):
        byte_string = b58decode(base58str)
        return cls(byte_string)

    def __init__(self, multihash):
        self.multihash = bytes(multihash)

    def __repr__(self):
        return u'MultihashReference: ' + self.multihash_base58()

    def __str__(self):
        return self.multihash_base58()

    def to_map(self):
        return {u'@link': self.multihash}

    def to_cbor_bytes(self):
        return cbor.dumps(self.to_map(), sort_keys=True)

    def multihash_base58(self):
        return b58encode(self.multihash)



#!/usr/bin/env python
import os, sys
from setuptools import setup, find_packages
from setuptools.command.build_py import build_py as _build_py

reqs_file = os.path.join(os.path.dirname(os.path.realpath(__file__))
                   , "requirements.txt")

reqs = None
with open(reqs_file) as f:
    reqs = f.readlines()

def _pre_build_py(dir):
    from subprocess import check_call
    check_call(['scripts/build_grpc.sh', sys.executable],
            cwd=dir)

class build_py(_build_py):
    def run(self):
        self.execute(_pre_build_py, [os.path.dirname(__file__)],
                     msg="Generating protobuf")
        _build_py.run(self)

setup(
    version='0.1.5',
    name='mediachain-client',
    description='mediachain reader command line interface',
    author='Mediachain Labs',
    packages=find_packages('.'),
    package_data={
        '': ['*.json']
    },
    author_email = 'hello@mediachainlabs.com',
    entry_points={
        'console_scripts': [
            'mediachain = mediachain.cli.main:main'
        ]
    },
    url='http://mediachain.io',
    install_requires=reqs,
    cmdclass={'build_py': build_py},
    setup_requires=['pytest-runner>=2.8'],
    tests_require=['pytest>=2.9.2'],
)



from mediachain.datastore import get_db, get_raw_datastore
from mediachain.ingestion.asset_loader import make_jpeg_data_uri
import copy
import base58
import json
from pygments import highlight
from utils import dump

def get_and_print_object(transactor, object_id):
    obj = get_object(transactor, object_id, fetch_images=False)
    dump(obj)


def get_object(transactor, object_id, fetch_images=True):
    db = get_db()
    base = db.get(object_id)
    head = transactor.get_chain_head(object_id)
    chain = get_object_chain(head, [])
    obj = reduce(chain_folder, chain, base)

    try:
        entity_id = obj['entity']
        obj['entity'] = get_object(transactor, entity_id, fetch_images)
    except KeyError as e:
        pass

    if fetch_images and obj.get('type') == 'artefact':
        return fetch_thumbnails(obj)
    return obj


def fetch_thumbnails(obj):
    def with_fallback():
        o = copy.deepcopy(obj)
        if 'meta' not in o:
            o['meta'] = {}
        if 'data' not in o['meta']:
            o['meta']['data'] = {}

        o['meta']['data']['thumbnail_base64'] = 'NO_IMAGE'
        return o

    try:
        thumb_ref_bytes = obj['meta']['data']['thumbnail']['@link']
        thumb_ref = base58.b58encode(thumb_ref_bytes)
        db = get_raw_datastore()
        thumb = db.get(thumb_ref)
    except (ValueError, LookupError):
        return with_fallback()

    with open('/tmp/thumbnail.jpg', 'wb') as f:
        f.write(thumb)

    with_thumb = copy.deepcopy(obj)
    with_thumb['meta']['data']['thumbnail_base64'] = make_jpeg_data_uri(thumb)
    return with_thumb


def apply_update_cell(acc, cell):
    result = copy.deepcopy(acc)
    cell = copy.deepcopy(cell)

    for k, v in cell['meta'].iteritems():
        result['meta'][k] = v

    return result

def apply_creation_cell(acc, update):
    result = copy.deepcopy(acc)

    try:
        result['entity'] = base58.b58encode(update['entity']['@link'])
    except KeyError as e:
        pass

    return result

def chain_folder(acc, x):
    cell_type = x.get('type')

    fn_map = {
        u'artefactCreatedBy': apply_creation_cell,
        u'artefactUpdate': apply_update_cell,
        u'entityUpdate': apply_update_cell
    }

    try:
        # apply a transform if we have one
        return fn_map[cell_type](acc, x)
    except KeyError as e:
        # otherwise, skip
        return acc

def get_object_chain(reference, acc):
    if reference is None:
        return acc

    db = get_db()
    obj = db.get(reference)

    try:
        next_ref = obj['chain']['@link']
        next_ref = base58.b58encode(next_ref)
    except KeyError as e:
        next_ref = None

    return get_object_chain(next_ref, acc + [obj])



import os
from dataset_iterator import DatasetIterator


class LocalFileIterator(DatasetIterator):
    def __init__(self, translator, root_path, limit=None):
        super(LocalFileIterator, self).__init__(translator, limit=limit)
        if not os.path.exists(root_path):
            raise ValueError('File at path {} does not exist'.format(root_path))
        
        self.root_path = root_path

    def generator(self):
        nn = 0
        limit = self.limit

        fns = []
        if os.path.isdir(self.root_path):
            for dir_name, subdir_list, file_list in os.walk(self.root_path):
                for fn in file_list:
                    fn = os.path.join(dir_name, fn)
                    if self.translator.can_translate_file(fn):
                        fns.append(fn)
        else:
            if self.translator.can_translate_file(self.root_path):
                fns.append(self.root_path)

        for fn in fns:
            if limit and (nn + 1 >= limit):
                print('ingested {} records, ending'.format(nn))
                return

            with open(fn, mode='rb') as f:
                try:
                    content = f.read()
                except IOError as e:
                    print('error reading from {}: {}'.format(
                        fn, str(e)
                    ))
                    continue

            nn += 1
            parsed = self.translator.parse(content)
            translated = self.translator.translate(parsed)
            local_assets = self.get_local_assets(fn, parsed)

            yield {
                'translated': translated,
                'raw_content': content,
                'parsed': parsed,
                'local_assets': local_assets
            }

    @staticmethod
    def get_local_assets(metadata_filepath, parsed_metadata):
        return {}
